-- regexpr.e version 1.0
-- David Cuny
-- dcuny@lanset.com

sequence 
    regExpr,    -- expression to parse
    matchChar   -- holds chars that match current pattern
    
integer 
    reps,       -- amount of times expr repeats
    exprAt      -- current location in regExpr

-----------------------------------------------------------------------------
procedure err( sequence text, sequence vals )
    printf( 1, text, vals )
    abort(0)
end procedure

-----------------------------------------------------------------------------
procedure advance()
    -- move ahead in expression
    
    -- can advance?
    if length( regExpr ) = exprAt then
        err( "Unexpected end in regexpr '%s'", {regExpr} )
    end if
    
    -- advance  
    exprAt = exprAt + 1
        
end procedure    


-----------------------------------------------------------------------------
procedure parseRegExpr()

    -- parse the char in regExpr

    integer low, high

    -- clear matched chars
    matchChar = repeat( 0, 256 )
    
    -- range?
    if regExpr[exprAt] = '[' then            

        -- move ahead
        advance()    
        
        while regExpr[exprAt] != ']' do
            
            -- read value
            low = regExpr[exprAt]
            advance()
            
            -- range?
            if regExpr[exprAt] = '-' then
                -- eat '-'
                advance()
                
                -- read high value
                high = regExpr[exprAt]
                advance()
                
                -- backwards?
                if low > high then
                    err( "Range '%s-%s' in '%s' is reversed",
                            {low, high, regExpr} )
                end if
                
                -- set values
                for i = low to high do
                    matchChar[i] = 1
                end for
                
            else
                -- just a value
                matchChar[low] = 1

            end if

        end while

        -- eat ']'
        advance()

        -- reps marks end of expr
        if regExpr[exprAt] = '?'
        or regExpr[exprAt] = '+'
        or regExpr[exprAt] = '*' then               
            -- save
            reps = regExpr[exprAt]
            advance()
        else
            err( "Expected '?', '*' or '+' following block '%s'",
                    {regExpr[exprAt]} )                
        end if


    else
        -- single value
        matchChar[regExpr[exprAt]] = 1
        advance()

        -- required
        reps = '+'
    
    end if

end procedure

-----------------------------------------------------------------------------
global function matchExpr( sequence expr, sequence text )

    -- attempt to match text against regExpr
    
    integer count, textAt
    
    -- initialize values
    regExpr = expr & 0
    text = text & 0
    exprAt = 1
    textAt = 1
    
    while 1 do         
    
        -- at end of expression?
        if exprAt = length( regExpr ) then
            -- matched
            return textAt
        end if
    
        -- get a part of the regExpr
        parseRegExpr()
        
        -- clear count
        count = 0

        -- try to match if not at end of text
        if textAt < length( text ) then      
        
            while matchChar[text[textAt]] do
            
                -- advance
                count = count + 1
                textAt = textAt + 1
                
                -- only match one char?
                if reps = '?' then
                    exit
                end if
                
                -- at end of text?
                if textAt = length( text ) then
                    -- exit
                    exit
                end if
                
            end while
        end if
        
        -- was there a minimum number of matches?
        if  reps = '+'
        and count = 0 then
            -- at least one match required
                return 0
        end if
        
    end while
    
end function        


